"""Trim whitespace inside of attributes where possible."""

import re

from ..preprocessors import BasePreprocessor
from ..rules import Rule
from ..types import Error
from . import BaseAttributeProcessor

# Match "..." or '...' with backslash escapes (no multiline strings)
STRING_RE = re.compile(
    r"""(?sx)
    ("(?:\\.|[^"\\])*")     # double-quoted string
  | ('(?:\\.|[^'\\])*')     # single-quoted string
""",
)

# 2+ horizontal whitespace, not touching \n, between non-whitespace chars
MIDDLE_WS_RUN = re.compile(r"(?<=\S)[^\S\n]{2,}(?=\S)")

# --- Encoded/escaped quote detection ----------------------------------------

# Safe encodings for a double quote (") that should NOT trigger a bail:
DOUBLE_QUOTE_SAFE_RE = re.compile(r"&quot;;?|&#34;;?|&#x22;;?|%22|\\u0022|\\x22", re.IGNORECASE)

# Safe encodings for a single quote / apostrophe (') that should NOT trigger a bail:
SINGLE_QUOTE_SAFE_RE = re.compile(r"&apos;;?|&#39;;?|&#x27;;?|%27|\\u0027|\\x27", re.IGNORECASE)


def has_inner_raw_bounding_quote(attr_body: str, bounding_character: str) -> bool:
    r"""Return True if attr_body contains a quote matching bounding_character.

    We first remove known *safe* encodings of that quote (entities, %XX,
    \uXXXX, \xXX) and then look for the literal quote. Backslash-escape
    (e.g. \") is NOT considered safe in HTML and will still contain a
    raw quote, so it will trigger a bail.
    """
    if bounding_character == '"':
        # Strip safe representations of double quotes so they don't confuse the check.
        stripped = DOUBLE_QUOTE_SAFE_RE.sub("", attr_body)
        return '"' in stripped
    if bounding_character == "'":
        stripped = SINGLE_QUOTE_SAFE_RE.sub("", attr_body)
        return "'" in stripped
    # Unexpected bounding char; be conservative.
    return True


def collapse_whitespace_outside_strings(string: str) -> str:
    """Collapse runs of 2+ horizontal whitespace to single spaces.

    Preserve whitespace inside string literals.
    """
    out = []
    pos = 0
    for match in STRING_RE.finditer(string):
        # process the gap before the string
        chunk = string[pos : match.start()]
        chunk = MIDDLE_WS_RUN.sub(" ", chunk)  # collapse to a single space
        out.append(chunk)

        # keep the string literal untouched
        out.append(match.group(0))
        pos = match.end()

    # tail after the last string
    tail = string[pos:]
    tail = MIDDLE_WS_RUN.sub(" ", tail)
    out.append(tail)

    return "".join(out)


class AttributeProcessor(BaseAttributeProcessor):
    """Trim whitespace inside of attributes where possible."""

    def process(
        self,
        attr_name: str,
        position: tuple[int, int],
        indentation: str,
        current_indentation_level: int,
        tab_width: int,
        line_length: int,
        max_items_per_line: int,
        bounding_character: str,
        preprocessor: BasePreprocessor | None,
        attr_body: str,
        *,
        solo: bool = False,
    ) -> tuple[str, list[Error]]:
        """Trim excess whitespace, and adjust starting & ending whitespace.

        If we detect a raw inner quote that matches the
        bounding_character, we bail (return the original body unchanged)
        to avoid breaking the HTML.
        """
        self._errors: list[Error] = []
        self.position = position

        if solo:
            brackets_quotes_and_space_length = 5
            extra_char_length = brackets_quotes_and_space_length
            extra_char_length += len(attr_name)
            extra_char_length += tab_width * current_indentation_level
            max_first_line_length = line_length - extra_char_length

        # --- Safety check: bail if an inner raw quote matches the bounding char
        if has_inner_raw_bounding_quote(attr_body, bounding_character):
            rule_code = "F16"
            error = Error(
                line=position[0],
                column=position[1],
                rule=Rule.get(rule_code),
                replacements={"attr": attr_name},
            )
            self._errors.append(error)

            replacement = {'"': "&quot;", "'": "&apos;"}[bounding_character]
            attr_body = attr_body.replace(bounding_character, replacement)

        # Collapse 2+ middle-of-line spaces/tabs *outside* strings
        adjusted_body = collapse_whitespace_outside_strings(attr_body)

        must_wrap = len(adjusted_body.strip()) > max_first_line_length if solo else False
        multiline = must_wrap or "\n" in adjusted_body.strip()

        if multiline:
            if re.match(r"^[^\S\n]*\S", attr_body):
                adjusted_body = adjusted_body.strip()
                # The first line is on the same line as the opening quotes.
                # Trim the leading whitespace:
                adjusted_body = re.sub(r"^[^\S\n]*(\S)", r"\1", adjusted_body)

                # Trim the trailing whitespace, including newlines:
                adjusted_body = re.sub(r"\s*$", "", adjusted_body)

            else:
                # Replace the end string with the proper spaced newline
                end_string = "".join(("\n", indentation * current_indentation_level))
                adjusted_body = re.sub(r"(\S)\s*$", rf"\1{end_string}", adjusted_body)

            # Replace triple-or-more newlines with double newlines
            adjusted_body = re.sub(r"(?:[^\S\n]*\r?\n){2,}", r"\n\n", adjusted_body)

            # Replace first-line and last-line trailing newlines
            adjusted_body = re.sub(r"^(?:\n[^\S\n]*){2,}", r"\n\n", adjusted_body)
            adjusted_body = re.sub(r"(?:[^\S\n]*\n)+([^\S\n]*)$", r"\n\1", adjusted_body)

            # Trim trailing whitespace on each line
            adjusted_body = re.sub(r"[^\S\n]+\n", r"\n", adjusted_body)
        else:
            # The only newlines were before or after all of the content. Trim them off.
            adjusted_body = re.sub(r"^\s*\n\s*", " ", adjusted_body)
            adjusted_body = re.sub(r"\s*\n\s*$", " ", adjusted_body)
            adjusted_body = adjusted_body.strip()

        return adjusted_body, self._errors

    def _handle_error(self, rule_code: str, **replacements: str) -> None:
        line, column = self.position
        self._errors.append(
            Error(
                line=line,
                column=column,
                rule=Rule.get(rule_code),
                replacements=replacements,
            ),
        )



"""Test the Tailwind attribute processor."""

import pytest

from cutesy.attribute_processors.class_ordering.tailwind import (
    AttributeProcessor,
    TailwindClass,
    group_and_sort,
    parse_tailwind_class,
)


class TestTailwind:
    """Test the Tailwind attribute processor."""

    @pytest.mark.parametrize(
        ("raw", "expected"),
        [
            # No modifiers
            (
                "bg-red-500",
                TailwindClass(
                    class_name="bg-red-500",
                    modifiers=[],
                    full_string="bg-red-500",
                ),
            ),
            # Single modifier
            (
                "hover:bg-red-500",
                TailwindClass(
                    class_name="bg-red-500",
                    modifiers=["hover"],
                    full_string="hover:bg-red-500",
                ),
            ),
            # Multiple modifiers
            (
                "md:hover:focus:text-lg",
                TailwindClass(
                    class_name="text-lg",
                    modifiers=["md", "hover", "focus"],
                    full_string="md:hover:focus:text-lg",
                ),
            ),
            # Arbitrary variant with brackets
            (
                "dark:[&>*]:text-center",
                TailwindClass(
                    class_name="text-center",
                    modifiers=["dark", "[&>*]"],
                    full_string="dark:[&>*]:text-center",
                ),
            ),
            # Escaped colon inside arbitrary value
            (
                r"content-\[url\:example\.com\]",
                TailwindClass(
                    class_name=r"content-\[url\:example\.com\]",
                    modifiers=[],
                    full_string=r"content-\[url\:example\.com\]",
                ),
            ),
            # Mixed responsive + state + complex arbitrary variant
            (
                r"rtl:focus:hover:[&>svg\:not(\.hidden)]:fill-current",
                TailwindClass(
                    class_name="fill-current",
                    modifiers=["rtl", "focus", "hover", r"[&>svg\:not(\.hidden)]"],
                    full_string=r"rtl:focus:hover:[&>svg\:not(\.hidden)]:fill-current",
                ),
            ),
            # Arbitrary value as base
            (
                r"text-\[theme\(colors\.blue-500\/70\%\)\]",
                TailwindClass(
                    class_name=r"text-\[theme\(colors\.blue-500\/70\%\)\]",
                    modifiers=[],
                    full_string=r"text-\[theme\(colors\.blue-500\/70\%\)\]",
                ),
            ),
            # Arbitrary property variant
            (
                r"[&\:where(h1,h2)]:underline",
                TailwindClass(
                    class_name="underline",
                    modifiers=[r"[&\:where(h1,h2)]"],
                    full_string=r"[&\:where(h1,h2)]:underline",
                ),
            ),
            # Class with dash prefix removed
            (
                "sm:-mt-4",
                TailwindClass(
                    class_name="mt-4",
                    modifiers=["sm"],
                    full_string="sm:-mt-4",
                ),
            ),
            # Container query with container declaration
            (
                "@container:bg-blue",
                TailwindClass(
                    class_name="bg-blue",
                    modifiers=["@container"],
                    full_string="@container:bg-blue",
                ),
            ),
            # Named container query modifier
            (
                "@lg:text-center",
                TailwindClass(
                    class_name="text-center",
                    modifiers=["@lg"],
                    full_string="@lg:text-center",
                ),
            ),
            # Combined regular and @ modifiers
            (
                "hover:@container:bg-red",
                TailwindClass(
                    class_name="bg-red",
                    modifiers=["hover", "@container"],
                    full_string="hover:@container:bg-red",
                ),
            ),
            # Container declaration standalone
            (
                "@container",
                TailwindClass(
                    class_name="@container",
                    modifiers=[],
                    full_string="@container",
                ),
            ),
        ],
    )
    def test_parse_tailwind_class(self, raw: str, expected: TailwindClass) -> None:
        """Test parsing a Tailwind class string."""
        got = parse_tailwind_class(raw)
        assert got == expected, f"\nRaw: {raw}\nExpected: {expected}\nGot: {got}"

    def test_modifiers_empty_when_no_modifiers(self) -> None:
        """Test modifiers are empty when no modifiers are present."""
        got = parse_tailwind_class("p-4")
        assert got.modifiers == []
        assert got.class_name == "p-4"
        assert got.full_string == "p-4"

    def test_keeps_backslashes_in_full_and_class_name(self) -> None:
        """Test backslashes are kept in full and class name."""
        raw = r"before:content-\['foo\:bar'\]"
        got = parse_tailwind_class(raw)
        assert got.modifiers == ["before"]
        assert got.class_name == r"content-\['foo\:bar'\]"
        assert got.full_string == raw


class TestAttributeProcessor:
    """Test the AttributeProcessor class."""

    def test_sort_with_grouped_false(self) -> None:
        """Test sort method with grouped=False returns flat list."""
        processor = AttributeProcessor()
        classes = ["bg-red-500", "text-white", "p-4"]
        result = processor.sort(classes, grouped=False)
        assert isinstance(result, list)
        assert all(isinstance(item, str) for item in result)

    def test_sort_with_grouped_true(self) -> None:
        """Test sort method with grouped=True returns list of lists."""
        processor = AttributeProcessor()
        classes = ["bg-red-500", "text-white", "p-4", "hover:bg-blue-500"]
        result = processor.sort(classes, grouped=True)
        assert isinstance(result, list)
        assert all(isinstance(group, list) for group in result)


class TestGroupAndSort:
    """Test the group_and_sort function."""

    def test_basic_grouping(self) -> None:
        """Test basic class grouping by category."""
        classes = ["bg-red-500", "p-4", "text-white", "flex"]
        groups = group_and_sort(classes)

        # Should have multiple groups
        assert len(groups) > 1

        # All groups should contain strings
        for group in groups:
            assert all(isinstance(tailwind_class, str) for tailwind_class in group)

    def test_user_defined_classes_last(self) -> None:
        """Test user-defined classes appear in the last group."""
        classes = ["bg-red-500", "my-custom-class", "p-4"]
        groups = group_and_sort(classes)

        # User-defined class should be in one of the groups (find it)
        found = False
        for group in groups:
            if "my-custom-class" in group:
                found = True
                break
        assert found

    def test_arbitrary_selector_variants(self) -> None:
        """Test arbitrary selector variants get their own groups."""
        classes = ["bg-red-500", "[&>*]:text-center", "p-4"]
        groups = group_and_sort(classes)

        # Should find the arbitrary selector class
        found = False
        for group in groups:
            if "[&>*]:text-center" in group:
                found = True
                break
        assert found

    def test_empty_input(self) -> None:
        """Test empty input returns empty list."""
        assert group_and_sort([]) == []

    def test_single_class(self) -> None:
        """Test single class input."""
        result = group_and_sort(["bg-red-500"])
        assert len(result) >= 1
        found = any("bg-red-500" in group for group in result)
        assert found

    def test_bem_notation_with_double_hyphens(self) -> None:
        """Test -- (double hyphens) is treated as user-defined."""
        classes = ["bg-red-500", "block__element--modifier", "p-4"]
        groups = group_and_sort(classes)

        # BEM class with -- should be treated as user-defined, not tailwind
        found_bem = False
        for group in groups:
            if "block__element--modifier" in group:
                found_bem = True
                break
        assert found_bem

    def test_bem_notation_with_double_underscores(self) -> None:
        """Test __ (double underscores) is treated as user-defined."""
        classes = ["bg-red-500", "block__element", "p-4"]
        groups = group_and_sort(classes)

        # BEM class with __ should be treated as user-defined, not tailwind
        found_bem = False
        for group in groups:
            if "block__element" in group:
                found_bem = True
                break
        assert found_bem

    def test_bem_notation_inside_arbitrary_values_allowed(self) -> None:
        """Test BEM notation inside arbitrary values [...] should be allowed.

        This covers the edge case where -- inside arbitrary values like
        mt-[var(--spacing)] should still be treated as valid Tailwind
        classes.
        """
        classes = ["bg-red-500", "mt-[var(--spacing)]", "p-4"]
        groups = group_and_sort(classes)

        # Class with -- inside arbitrary value should be treated as tailwind
        found_arbitrary = False
        for group in groups:
            if "mt-[var(--spacing)]" in group:
                found_arbitrary = True
                break
        assert found_arbitrary

    def test_prefix_removal_edge_cases(self) -> None:
        """Test edge cases for prefix removal in class name parsing."""
        # Test double dash prefix (like CSS custom properties)
        result = parse_tailwind_class("--custom-property")
        assert result.class_name == "-custom-property"  # Only first dash removed
        assert result.modifiers == []

        # Test single dash prefix
        result = parse_tailwind_class("-mt-4")
        assert result.class_name == "mt-4"  # Dash prefix removed
        assert result.modifiers == []

    def test_at_container_queries(self) -> None:
        """Test @ container queries are properly parsed and sorted."""
        classes = ["@lg:text-lg", "bg-red-500", "hover:@max-sm:p-4", "@container"]
        groups = group_and_sort(classes)

        # Should find classes with @ modifiers and declarations
        all_classes = [class_name for group in groups for class_name in group]
        assert "@lg:text-lg" in all_classes
        assert "hover:@max-sm:p-4" in all_classes
        assert "bg-red-500" in all_classes
        assert "@container" in all_classes



"""Tests for Django preprocessor functionality."""

import pytest

from cutesy.preprocessors.django import Preprocessor
from cutesy.types import InstructionType, StructuralError


class TestDjangoPreprocessor:
    """Test Django template preprocessor."""

    def test_initialization(self) -> None:
        """Test Django preprocessor initialization."""
        preprocessor = Preprocessor()

        # Check class variables are set correctly
        assert ("{%", "%}") in preprocessor.braces
        assert ("{{", "}}") in preprocessor.braces
        assert ("{#", "#}") in preprocessor.braces

        assert "freeform" in preprocessor.closing_tag_string_map
        assert "comment" in preprocessor.closing_tag_string_map

    def test_parse_instruction_tag_value_type(self) -> None:
        """Test parsing Django variable tags (value type)."""
        preprocessor = Preprocessor()
        html = "{{ variable }}"

        instruction, instruction_type = preprocessor.parse_instruction_tag(
            ("{{", "}}"),
            html,
            0,
            len(html) - 2,
        )

        assert instruction == "…"
        assert instruction_type == InstructionType.VALUE

    def test_parse_instruction_tag_if_conditional(self) -> None:
        """Test parsing Django if tag."""
        preprocessor = Preprocessor()
        html = "{% if condition %}"

        instruction, instruction_type = preprocessor.parse_instruction_tag(
            ("{%", "%}"),
            html,
            0,
            len(html) - 2,
        )

        assert instruction == "if"
        assert instruction_type == InstructionType.CONDITIONAL

    def test_parse_instruction_tag_endif_conditional(self) -> None:
        """Test parsing Django endif tag."""
        preprocessor = Preprocessor()
        html = "{% endif %}"

        instruction, instruction_type = preprocessor.parse_instruction_tag(
            ("{%", "%}"),
            html,
            0,
            len(html) - 2,
        )

        assert instruction == "endif"
        assert instruction_type == InstructionType.END_CONDITIONAL

    def test_parse_instruction_tag_elif_conditional(self) -> None:
        """Test parsing Django elif tag."""
        preprocessor = Preprocessor()
        html = "{% elif condition %}"

        instruction, instruction_type = preprocessor.parse_instruction_tag(
            ("{%", "%}"),
            html,
            0,
            len(html) - 2,
        )

        assert instruction == "elif"
        assert instruction_type == InstructionType.MID_CONDITIONAL

    def test_parse_instruction_tag_else_conditional(self) -> None:
        """Test parsing Django else tag."""
        preprocessor = Preprocessor()
        html = "{% else %}"

        instruction, instruction_type = preprocessor.parse_instruction_tag(
            ("{%", "%}"),
            html,
            0,
            len(html) - 2,
        )

        assert instruction == "else"
        assert instruction_type == InstructionType.LAST_CONDITIONAL

    def test_parse_instruction_tag_for_repeatable(self) -> None:
        """Test parsing Django for tag."""
        preprocessor = Preprocessor()
        html = "{% for item in items %}"

        instruction, instruction_type = preprocessor.parse_instruction_tag(
            ("{%", "%}"),
            html,
            0,
            len(html) - 2,
        )

        assert instruction == "for"
        assert instruction_type == InstructionType.REPEATABLE

    def test_parse_instruction_tag_endfor_repeatable(self) -> None:
        """Test parsing Django endfor tag."""
        preprocessor = Preprocessor()
        html = "{% endfor %}"

        instruction, instruction_type = preprocessor.parse_instruction_tag(
            ("{%", "%}"),
            html,
            0,
            len(html) - 2,
        )

        assert instruction == "endfor"
        assert instruction_type == InstructionType.END_REPEATABLE

    def test_parse_instruction_tag_empty_for(self) -> None:
        """Test parsing Django empty tag in for loops."""
        preprocessor = Preprocessor()
        html = "{% empty %}"

        instruction, instruction_type = preprocessor.parse_instruction_tag(
            ("{%", "%}"),
            html,
            0,
            len(html) - 2,
        )

        assert instruction == "empty"
        assert instruction_type == InstructionType.MID_CONDITIONAL

    def test_parse_instruction_tag_block_partial(self) -> None:
        """Test parsing Django block tag."""
        preprocessor = Preprocessor()
        html = "{% block content %}"

        instruction, instruction_type = preprocessor.parse_instruction_tag(
            ("{%", "%}"),
            html,
            0,
            len(html) - 2,
        )

        assert instruction == "block"
        assert instruction_type == InstructionType.PARTIAL

    def test_parse_instruction_tag_endblock_partial(self) -> None:
        """Test parsing Django endblock tag."""
        preprocessor = Preprocessor()
        html = "{% endblock %}"

        instruction, instruction_type = preprocessor.parse_instruction_tag(
            ("{%", "%}"),
            html,
            0,
            len(html) - 2,
        )

        assert instruction == "endblock"
        assert instruction_type == InstructionType.END_PARTIAL

    def test_parse_instruction_tag_with_partial(self) -> None:
        """Test parsing Django with tag."""
        preprocessor = Preprocessor()
        html = "{% with value=variable %}"

        instruction, instruction_type = preprocessor.parse_instruction_tag(
            ("{%", "%}"),
            html,
            0,
            len(html) - 2,
        )

        assert instruction == "with"
        assert instruction_type == InstructionType.PARTIAL

    def test_parse_instruction_tag_endwith_partial(self) -> None:
        """Test parsing Django endwith tag."""
        preprocessor = Preprocessor()
        html = "{% endwith %}"

        instruction, instruction_type = preprocessor.parse_instruction_tag(
            ("{%", "%}"),
            html,
            0,
            len(html) - 2,
        )

        assert instruction == "endwith"
        assert instruction_type == InstructionType.END_PARTIAL

    def test_parse_instruction_tag_comment_freeform(self) -> None:
        """Test parsing Django comment tag."""
        preprocessor = Preprocessor()
        html = "{% comment %}"

        instruction, instruction_type = preprocessor.parse_instruction_tag(
            ("{%", "%}"),
            html,
            0,
            len(html) - 2,
        )

        assert instruction == "comment"
        assert instruction_type == InstructionType.COMMENT

    def test_parse_instruction_tag_endcomment_freeform(self) -> None:
        """Test parsing Django endcomment tag."""
        preprocessor = Preprocessor()
        html = "{% endcomment %}"

        instruction, instruction_type = preprocessor.parse_instruction_tag(
            ("{%", "%}"),
            html,
            0,
            len(html) - 2,
        )

        assert instruction == "endcomment"
        assert instruction_type == InstructionType.END_COMMENT

    def test_parse_instruction_tag_spaceless_freeform(self) -> None:
        """Test parsing Django spaceless tag."""
        preprocessor = Preprocessor()
        html = "{% spaceless %}"

        instruction, instruction_type = preprocessor.parse_instruction_tag(
            ("{%", "%}"),
            html,
            0,
            len(html) - 2,
        )

        assert instruction == "spaceless"
        assert instruction_type == InstructionType.FREEFORM

    def test_parse_instruction_tag_endspaceless_freeform(self) -> None:
        """Test parsing Django endspaceless tag."""
        preprocessor = Preprocessor()
        html = "{% endspaceless %}"

        instruction, instruction_type = preprocessor.parse_instruction_tag(
            ("{%", "%}"),
            html,
            0,
            len(html) - 2,
        )

        assert instruction == "endspaceless"
        assert instruction_type == InstructionType.END_FREEFORM

    def test_parse_instruction_tag_blocktrans_conditional(self) -> None:
        """Test parsing Django blocktrans tag."""
        preprocessor = Preprocessor()
        html = "{% blocktrans %}"

        instruction, instruction_type = preprocessor.parse_instruction_tag(
            ("{%", "%}"),
            html,
            0,
            len(html) - 2,
        )

        assert instruction == "blocktrans"
        assert instruction_type == InstructionType.CONDITIONAL

    def test_parse_instruction_tag_plural_conditional(self) -> None:
        """Test parsing Django plural tag."""
        preprocessor = Preprocessor()
        html = "{% plural %}"

        instruction, instruction_type = preprocessor.parse_instruction_tag(
            ("{%", "%}"),
            html,
            0,
            len(html) - 2,
        )

        assert instruction == "plural"
        assert instruction_type == InstructionType.LAST_CONDITIONAL

    def test_parse_instruction_tag_endblocktrans_conditional(self) -> None:
        """Test parsing Django endblocktrans tag."""
        preprocessor = Preprocessor()
        html = "{% endblocktrans %}"

        instruction, instruction_type = preprocessor.parse_instruction_tag(
            ("{%", "%}"),
            html,
            0,
            len(html) - 2,
        )

        assert instruction == "endblocktrans"
        assert instruction_type == InstructionType.END_CONDITIONAL

    def test_parse_instruction_tag_while_repeatable(self) -> None:
        """Test parsing Django while tag."""
        preprocessor = Preprocessor()
        html = "{% while condition %}"

        instruction, instruction_type = preprocessor.parse_instruction_tag(
            ("{%", "%}"),
            html,
            0,
            len(html) - 2,
        )

        assert instruction == "while"
        assert instruction_type == InstructionType.REPEATABLE

    def test_parse_instruction_tag_endwhile_repeatable(self) -> None:
        """Test parsing Django endwhile tag."""
        preprocessor = Preprocessor()
        html = "{% endwhile %}"

        instruction, instruction_type = preprocessor.parse_instruction_tag(
            ("{%", "%}"),
            html,
            0,
            len(html) - 2,
        )

        assert instruction == "endwhile"
        assert instruction_type == InstructionType.END_REPEATABLE

    def test_parse_instruction_tag_comment_ignored(self) -> None:
        """Test parsing Django comment tags."""
        preprocessor = Preprocessor()
        html = "{# This is a comment #}"

        instruction, instruction_type = preprocessor.parse_instruction_tag(
            ("{#", "#}"),
            html,
            0,
            len(html) - 2,
        )

        assert instruction == "…"
        assert instruction_type == InstructionType.IGNORED

    def test_parse_instruction_tag_freeform_special_comment(self) -> None:
        """Test parsing special freeform comment directive."""
        preprocessor = Preprocessor()
        html = "{# freeform #}"

        instruction, instruction_type = preprocessor.parse_instruction_tag(
            ("{#", "#}"),
            html,
            0,
            len(html) - 2,
        )

        assert instruction == "freeform"
        assert instruction_type == InstructionType.FREEFORM

    def test_parse_instruction_tag_endfreeform_special_comment(self) -> None:
        """Test parsing special endfreeform comment directive."""
        preprocessor = Preprocessor()
        html = "{# endfreeform #}"

        instruction, instruction_type = preprocessor.parse_instruction_tag(
            ("{#", "#}"),
            html,
            0,
            len(html) - 2,
        )

        assert instruction == "endfreeform"
        assert instruction_type == InstructionType.END_FREEFORM

    def test_parse_instruction_tag_empty_comment(self) -> None:
        """Test parsing empty Django comment."""
        preprocessor = Preprocessor()
        html = "{# #}"

        instruction, instruction_type = preprocessor.parse_instruction_tag(
            ("{#", "#}"),
            html,
            0,
            len(html) - 2,
        )

        assert instruction == "…"
        assert instruction_type == InstructionType.IGNORED

    def test_parse_instruction_tag_unknown_tag_as_value(self) -> None:
        """Test parsing unknown Django tag defaults to value type."""
        preprocessor = Preprocessor()
        html = "{% unknown_tag %}"

        instruction, instruction_type = preprocessor.parse_instruction_tag(
            ("{%", "%}"),
            html,
            0,
            len(html) - 2,
        )

        assert instruction == "unknown_tag"
        assert instruction_type == InstructionType.VALUE

    def test_parse_instruction_tag_empty_tag_raises_error(self) -> None:
        """Test parsing empty Django tag raises error."""
        preprocessor = Preprocessor()
        html = "{% %}"

        # Initialize the attributes that make_fatal_error expects
        preprocessor.line = 1
        preprocessor.offset = 0

        # This should raise a StructuralError for P4 code
        with pytest.raises(StructuralError) as exc_info:
            preprocessor.parse_instruction_tag(("{%", "%}"), html, 0, len(html) - 2)

        assert exc_info.value.errors[0].rule.code == "P4"
